/*
    dltrace v0.5
    Copyright (C) 2005 Richard Johnson <rjohnson@idefense.com>

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 59 Temple
    Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <asm/page.h>
// asm/page.h -- linux 
// i386/vmparam.h -- nbsd 
// i386/param.h -- obsd 
#include "melf.h"
#include "plump.h"
#include "lists.h"





/******************************************************************************
** data types
*/

typedef struct library {
	uint8_t  *path;
	uint32_t  addr;
	uint32_t  hash;
	list_t   *symbols;  // data symbols 
	MELF     *melf;	
} library_t;

typedef struct process {	
	uint32_t pid;           
	char   **argv;         
	uint32_t addr;
	uint32_t e_entry;
	uint32_t textsz;
	uint8_t hash;
	library_t *interp; 

	plumpreg_t regs[15];
	uint32_t depth;
	uint32_t last_call;

	splay_t *local_functions;  // local function symbols
	splay_t *shlib_functions;  // shared library calls for ltrace
	list_t  *callstack;  // stack of return breakpoints
	list_t  *mmaps;
	list_t  *liblist;       // list of librarys 
	list_t  *symbols;
	MELF    *melf;
	plump_t *plump;
} proc_t;

typedef struct symbol {
	int8_t *name;
	uint32_t  addr;
	uint32_t  size;
	uint32_t  argc;
	uint32_t  analyzed;
	library_t    *library;
} symbol_t;	

typedef struct breakpoint {
	uint32_t addr;
	symbol_t *symbol; // related symbol
} breakpoint_t, callstack_t;		


#define MAX_CALLDEPTH 1024
struct options {
	int32_t attach;
	int32_t ltrace;
	int32_t strace;
	int32_t btrace;
	int32_t depth;	 // option 'd'
	int32_t regs;
	int32_t tree;
	int32_t verbose;           
	FILE   *logfile; // option 'o'
	FILE   *xmlfile; // option 'o'
};


/******************************************************************************
** globals 
*/

struct options options;
list_t  *pslist;


/******************************************************************************
** macro defines
*/

#define ELFSIG    0x464c457f   // integer value for matching "\x7fELF"

#define HASH_DATASZ  4096
#define HASH_KEY  0x1d3f1ab5



#define EAX (uint32_t)ps->regs[0].val.b32
#define EBX (uint32_t)ps->regs[1].val.b32
#define ECX (uint32_t)ps->regs[2].val.b32
#define EDX (uint32_t)ps->regs[3].val.b32
#define EDI (uint32_t)ps->regs[4].val.b32
#define ESI (uint32_t)ps->regs[5].val.b32
#define ESP (uint32_t)ps->regs[6].val.b32
#define EBP (uint32_t)ps->regs[7].val.b32
#define EIP (uint32_t)ps->regs[8].val.b32
#define SS (uint16_t)ps->regs[9].val.b16
#define CS (uint16_t)ps->regs[10].val.b16
#define DS (uint16_t)ps->regs[11].val.b16
#define ES (uint16_t)ps->regs[12].val.b16
#define FS (uint16_t)ps->regs[13].val.b16
#define GS (uint16_t)ps->regs[14].val.b16


/******************************************************************************
** functions
*/

// elf.c
int32_t   dt_load_binary(proc_t *ps);
list_t  * dt_load_symbols(MELF *melf, proc_t *ps);
int32_t   dt_load_interp(MELF *melf, proc_t *ps);
int32_t   dt_load_dynamic(MELF *melf, proc_t *ps, uint32_t depth);
int32_t   dt_load_library(proc_t *ps, uint8_t *path, uint32_t depth);
uint32_t  dt_elf_hash_file(uint8_t *path);
int32_t   dt_disas(uint8_t *data, uint32_t size);
int32_t   dt_load_syscalls(proc_t *ps, uint8_t *path);

// proc.c
int32_t   dt_open_process(proc_t *ps);
int32_t   dt_library_reloc_syms(proc_t *ps);
uint8_t * dt_mem_get_chunk(proc_t *ps, uint32_t addr, uint32_t len);
uint8_t * dt_mem_get_string(proc_t *ps, uint32_t addr, uint32_t len);

// trace.c
int32_t   dt_init_trace(proc_t *ps);
int32_t   dt_init_ltrace(proc_t *ps);
int32_t   dt_run_trace(proc_t *ps);
int32_t   dt_run_ltrace(proc_t *ps); 
int32_t   dt_function_analyze(proc_t *ps, uint32_t addr, uint32_t len);
void      dt_print_ltrace_ret(proc_t *ps, callstack_t *bp, uint32_t retaddr);
void      dt_print_ltrace_call(proc_t *ps, callstack_t *bp, uint32_t retaddr);

// hash.c
uint32_t  dt_hash_calc (unsigned char *key, unsigned long length, unsigned long initval);

void      vlog(FILE *fd, const char *fmt, ...);
int32_t   sym_find_by_addr(void *, void *);
int32_t   sym_find_by_name(void *, void *);


/*
int dt_breakpoint_add(proc_t *ps, splay_t *tree, breakpoint_t *bp);
struct symbol dt_resolv_addr(proc_t *ps); 
struct symbol dt_resolv_sym(proc_t *ps);
int vlog(FILE *fd, const char *fmt, ...);
int dt_library_reloc_syms(proc_t *ps);

uint8_t *dt_mem_get_chunk(proc_t *ps, uint32_t addr, uint32_t len);
uint8_t *dt_mem_get_string(proc_t *ps, uint32_t addr, uint32_t len);


int dt_function_analyze(proc_t *ps, uint32_t addr, uint32_t len);
int dt_open_process(proc_t *ps);
	int dt_load_binary(proc_t *ps);
		list_t * dt_load_symbols(MELF *melf, proc_t *ps);	
//		int dt_load_interp(MELF *melf, proc_t *ps);
		int dt_load_dynamic(MELF *melf, proc_t *ps);
		int dt_load_library(proc_t *ps, char *path, int load_addr);
			// setup library struct
			// recurse
int dt_init_trace(proc_t *ps); // set breakpoints
	int dt_init_ltrace(proc_t *ps);
int dt_run_trace(proc_t *ps);
	int dt_run_ltrace(proc_t *ps);
	int dt_handle_signal(proc_t *ps);
		int dt_sig_segv(proc_t *ps);
			// dump useful info in case of crashes 
*/	
