#include <stdlib.h>
#include <stdio.h>

#include "plump.h"

#define USER_BREAKPOINT 0x0804850c
#define USER_WRITELOC   0xbfffff00

char msg[] = "This is a test.";

int main(int argc, char **argv)
{
	int ret = 0, step = 0;
	plumpaddr_t stop = 0;
	char tmp[64];
	plump_t *p;

	if (argc == 1)
	{
		fprintf(stdout, "Usage: %s [pid]\n", argv[0]);
		return 0;
	}

	memset(tmp, 0, sizeof(tmp));

	do
	{
		// Step #1: Open
		fprintf(stdout, "(Step = %d): Opening...\n", ++step);

		if ((ret = plump_open(atoi(argv[1]), PLUMP_ACCESS_ALL, PLUMP_MEDIUM_ANY, &p)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_open() failed, ret=%d, step=%d.\n", ret, step);
			break;
		}

		// Step #2: Continue indefinitely
		fprintf(stdout, "(Step = %d): Continuing indefinitely...\n", ++step);

		if ((ret = plump_execContinue(p, 0, NULL, PLUMP_CONTINUE_FLAG_INDEFINITELY)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_execContinue() failed, ret=%d, step=%d\n", ret, step);
			break;
		}

		// Step #3: Stopping
		fprintf(stdout, "(Step = %d): Stopping...\n", ++step);

		if ((ret = plump_execStop(p)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_execStop() failed, ret=%d, step=%d\n", ret, step);
			break;
		}

		// Step #4: Continue until source breakpoint
		fprintf(stdout, "(Step = %d): Running until source bp...\n", ++step);

		if ((ret = plump_execContinue(p, 0, &stop, PLUMP_CONTINUE_FLAG_UNTILSRCBP)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_execContinue() failed, ret=%d, step=%d\n", ret, step);
			break;
		}
		else
			fprintf(stdout, "(Step = %d): Success => Hit source breakpoint at %.8x\n", step, (unsigned int)stop);

		// Reset stop.
		stop = 0;

		// Step #5: Add user breakpoint
		fprintf(stdout, "(Step = %d): Adding user bp...\n", ++step);

		if ((ret = plump_breakpointAdd(p, (plumpaddr_t)USER_BREAKPOINT)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_breakpointAdd() failed, ret=%d, step=%d\n", ret, step);
			break;
		}

		// Step #6: Run until user breakpoint
		fprintf(stdout, "(Step = %d): Running until user bp...\n", ++step);

		if ((ret = plump_execContinue(p, 0, &stop, PLUMP_CONTINUE_FLAG_UNTILUSRBP)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_execContinue() failed, ret=%d, step=%d\n", ret, step);
			break;
		}
		else
			fprintf(stdout, "(Step = %d): Success => Hit user breakpoint at %.8x (userbp=%.8x)\n", step, (unsigned int)stop, USER_BREAKPOINT);

		// Step #7: Remove user breakpoint
		fprintf(stdout, "(Step = %d): Removing user bp...\n", ++step);

		if ((ret = plump_breakpointRemove(p, (plumpaddr_t)USER_BREAKPOINT)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_breakpointRemove() failed, ret=%d, step=%d\n", ret, step);
			break;
		}

		// Step #8: Writing memory
		fprintf(stdout, "(Step = %d): Checking if %.8x is valid...\n", ++step, USER_WRITELOC);

		if ((ret = plump_memoryIsValid(p, (plumpaddr_t)USER_WRITELOC)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_memoryIsValid() failed, ret=%d, step=%d\n", ret, step);
			break;
		}
		
		// Step #9: Writing memory
		fprintf(stdout, "(Step = %d): Writing to %.8x...\n", ++step, USER_WRITELOC);

		if ((ret = plump_memoryWrite(p, (plumpaddr_t)USER_WRITELOC, msg, sizeof(msg), NULL)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_memoryIsValid() failed, ret=%d, step=%d\n", ret, step);
			break;
		}

		// Step #10: Reading memory
		fprintf(stdout, "(Step = %d): Writing to %.8x...\n", ++step, USER_WRITELOC);

		if ((ret = plump_memoryRead(p, (plumpaddr_t)USER_WRITELOC, tmp, sizeof(msg), NULL)) != PLUMP_SUCCESS)
		{
			fprintf(stdout, "plump_memoryIsValid() failed, ret=%d, step=%d\n", ret, step);
			break;
		}
		else
			fprintf(stdout, "(Step = %d): Success, read '%s'.\n", step, tmp);

	} while (0);

	if (ret >= PLUMP_SUCCESS)
		fprintf(stdout, "overall success!\n");
	else
		fprintf(stdout, "overall failure!\n");

	if (p)
		plump_close(p);

	return 1;
}
